from ganttlogger.modules.InitProcess import InitProcess
from ganttlogger.modules.Public import StrFormatter
from ganttlogger.modules.Alone import Alone
from ganttlogger.modules.Observer import WinObserver, MacObserver
from ganttlogger.modules.Logger import Logger
from ganttlogger.modules.Plotter import Plotter
from ganttlogger.modules.Displayer import Displayer

def main():
    '''
    Initial Process
    '''
    # A start of a module 'StrFormatter' for coloring terminal
    strformatter = StrFormatter()
    strformatter.start()
    # Main initialization
    init = InitProcess()
    os, mode, uuid = init.get_init_parameters()

    # Start main process(thread-loop) in accordance with mode
    if mode == "Alone":
        alone = Alone(os, uuid)
        alone.run()
    elif mode == "AloneWithPlot":
        alone = Alone(os, uuid, withplot=True)
        alone.run()
    elif mode == "Observer":
        print("We can't execute Observer because it hasn't been implemented.")
        # if os == "w":
        #     observer = WinObserver(uuid=uuid, is_alone=False)
        # elif os == "d":
        #     observer = MacObserver(uuid=uuid, is_alone=False)
        # observer.start()
    elif mode == "Logger":
        print("We can't execute Logger because it hasn't been implemented.")
        # logger = Logger(uuid)
        # plotter = Plotter(uuid)
        # logger.run_logger()
        # plotter.run()
    elif mode == "Plotter":
        plotter = Plotter()
        plotter.start()
    elif mode == "Displayer":
        displayer = Displayer()
        displayer.start()


if __name__ == "__main__":
    main()






__copyright__ = "Copyright (C) 2019 KagenoMoheji"
__version__ = '0.1.0'
__license__ = "MIT"
__author__ = "Goki Sugimura(KagenoMoheji)"
__author_email__ = 'shadowmoheji.pd@gmail.com'
__url__="https://github.com/KagenoMoheji/GanttLogger"


'''
Because we can run MouseObserver and KeyboardObserver with same code
regardless on Windows or Mac, we implemented common classes here.
But we can implement in per classes when we have to implement following OS.
'''
import math
import time
from datetime import datetime
from collections import deque
# import pyautogui
from pynput import mouse, keyboard
from ganttlogger.modules.Public import MyThread, StrFormatter
import ganttlogger.modules.Global as global_v

class MouseObserver:
    uuid = ""
    sec_sum_mouse_dist = 0
    mouse_ctrl = None
    data_process = None
    def __init__(self, uuid, is_alone):
        self.uuid = uuid
        self.sec_sum_mouse_dist = 0
        self.mouse_ctrl = mouse.Controller()
        if is_alone:
            self.data_process = self.enqueue_data
        else:
            self.data_process = self.send_json

    def run(self):
        try:
            dif_x, dif_y = 0, 0
            recent_time = time.time()
            recent_x, recent_y = self.mouse_ctrl.position # pyautogui.position()
            while not global_v.is_sleeping:
                try:
                    current_time = time.time()
                    x, y = self.mouse_ctrl.position # pyautogui.position()
                    if (x != recent_x) or (y != recent_y):
                        dif_x = x - recent_x
                        dif_y = y - recent_y
                        self.sec_sum_mouse_dist += math.sqrt(dif_x * dif_x + dif_y * dif_y)
                        recent_x, recent_y = x, y
                    if current_time - recent_time > 1.0:
                        self.data_process(datetime.now().strftime("%Y/%m/%d %H:%M:%S.%f"), self.sec_sum_mouse_dist)
                        # print("Mouse[{datetime}]: {dist}".format(datetime=datetime.now().strftime("%Y/%m/%d %H:%M:%S.%f"), dist=self.sec_sum_mouse_dist))
                        recent_time = current_time
                        self.sec_sum_mouse_dist = 0
                    time.sleep(0.001)
                except TypeError:
                    # Avoid to exit thread loop (by rebooting from pc sleep)
                    continue
        except:
            # If this thread stopped by rebooting from sleep, maybe...
            import traceback
            print("Thread loop exited by any problem!!!!")
            global_v.is_threadloop_error = True
            global_v.is_sleeping = True
            traceback.print_exc()
        # except KeyboardInterrupt:
        #     print("MouseObserver.py: KeyboardInterrupt")
    
    def send_json(self, t, dis_sec):
        pass

    def enqueue_data(self, t, dis_sec):
        global_v.mouse_id += 1
        global_v.mouse_queue.append({
            "uuid": self.uuid,
            "type": "m",
            "id": global_v.mouse_id,
            "distance": dis_sec,
            "time": t
        })


class KeyboardObserver:
    uuid = ""
    data_process = None
    sec_sum_keyboard_cnt = 0
    current_4key = deque([], maxlen=4)
    EXITCOMB_WIN = set([
        keyboard.Key.ctrl_l,
        keyboard.Key.shift_l,
        keyboard.Key.ctrl_r,
        keyboard.Key.shift_r
        # keyboard.KeyCode(char = ''),
        # keyboard.KeyCode(char = 'q')
    ])
    EXITCOMB_MAC = set([
        keyboard.Key.cmd_l,
        keyboard.Key.shift_l,
        keyboard.Key.cmd_r,
        keyboard.Key.shift_r
    ])
    def __init__(self, uuid, is_alone):
        self.uuid = uuid
        self.sec_sum_keyboard_cnt = 0
        if is_alone:
            self.data_process = self.enqueue_data
        else:
            self.data_process = self.send_json

    def on_release(self, key):
        if len(self.current_4key) == 4:
            self.current_4key.popleft()
            self.current_4key.append(key)
        else:
            self.current_4key.append(key)
        if (self.EXITCOMB_WIN == set(self.current_4key)) or (self.EXITCOMB_MAC == set(self.current_4key)):
            # Here, switch a flag to exit children threads
            global_v.is_sleeping = True
            '''
            ###############################################################
            Delete all text in terminal(Flush stdin buffer?)
            ??????????????????????????????????????????
            ###############################################################
            '''
            return False
        self.sec_sum_keyboard_cnt += 1
        time.sleep(0.001) # Max type speed is 256 wpm -> 0.002 is OK?
        return True
    
    def get_key_second(self):
        try:
            while not global_v.is_sleeping:
                self.data_process(datetime.now().strftime("%Y/%m/%d %H:%M:%S.%f"), self.sec_sum_keyboard_cnt)
                # print("Keyboard[{datetime}]: {cnt}".format(datetime=datetime.now().strftime("%Y/%m/%d %H:%M:%S.%f"), cnt=self.sec_sum_keyboard_cnt))
                print("""\
============[tab]==============\n
{t}
-----------[mouse]-----------\n
{m}
-----------[keyboard]-----------\n
{k}
""".format(t=global_v.tab_queue, m=global_v.mouse_queue, k=global_v.keyboard_queue))
                self.sec_sum_keyboard_cnt = 0
                time.sleep(1)
        except:
            # If this thread stopped by rebooting from sleep, maybe...
            import traceback
            print("Thread loop exited by any problem!!!!")
            global_v.is_threadloop_error = True
            global_v.is_sleeping = True
            traceback.print_exc()
    
    def run(self):
        try:
            listener = keyboard.Listener(on_release=self.on_release)
            th_on_release = MyThread(target=listener.start)
            th_mainloop = MyThread(target=self.get_key_second)
            th_on_release.start()
            th_mainloop.start()
        # except KeyboardInterrupt:
        #     print("KeyboardObserver.py: KeyboardInterrupt")
        finally:
            th_on_release.stop()
            th_mainloop.stop()

    def send_json(self, t, cnt_sec):
        pass

    def enqueue_data(self, t, cnt_sec):
        global_v.keyboard_id += 1
        global_v.keyboard_queue.append({
            "uuid": self.uuid,
            "type": "k",
            "id": global_v.keyboard_id,
            "count": cnt_sec,
            "time": t
        })


import sys
import time
from datetime import datetime
# import numpy as np
import concurrent.futures as confu
from collections import deque
import ganttlogger.modules.Global as global_v
from ganttlogger.modules.Public import StrFormatter
from ganttlogger.modules.CommonObservePackages import MouseObserver, KeyboardObserver

class Observer:
    strfmr = None
    ob_activetab = None
    ob_mouse = None
    ob_keyboard = None
    executor = None
    def __init__(self, uuid, is_alone):
        '''
        Monitor starting time of active-tab, count of keyboard, distance of mouse(cursor)-moving,
        and enqueue these data.
        '''
        self.ob_mouse = MouseObserver(uuid, is_alone)
        self.ob_keyboard = KeyboardObserver(uuid, is_alone)
        self.strfmr = StrFormatter()
        self.executor = confu.ThreadPoolExecutor(max_workers=3)

    def start(self):
        print(self.strfmr.get_colored_console_log("yellow",
            "Start after finising setting logger.\nReady to start?(Y/n) : "), end="")
        st_input = input().strip()
        if st_input == "n":
            print("GanttLogger closed")
            sys.exit()
        if st_input != "Y":
            print(self.strfmr.get_colored_console_log("red",
                "Error: Invalid input. Input 'Y'(=yes) or 'n'(=no)."))
            sys.exit()
        self.run()

    def run(self):
        while True:
            self.executor.submit(self.ob_activetab.run)
            self.executor.submit(self.ob_mouse.run)
            self.executor.submit(self.ob_keyboard.run)
            while not global_v.is_sleeping:
                time.sleep(1)
            if global_v.is_threadloop_error: # When "Thread loop exited by any problem!!!!" occured
                self.executor.shutdown()
                sys.exit()
            is_confirmed_exiting = self.confirm_exiting()
            if is_confirmed_exiting:
                print(self.strfmr.get_colored_console_log("yellow",
                    "Observer exited."))
                break
            else:
                print(self.strfmr.get_colored_console_log("yellow",
                    "Observer restarted."))
                global_v.is_sleeping = False
        self.executor.shutdown()
        global_v.cli_exit = True

    def confirm_exiting(self):
        while True:
            print(self.strfmr.get_colored_console_log("yellow",
                "Logging is sleeping. Will you exit?(Y/n) : "), end="")
            str_input = input().strip()
            if str_input == "Y":
                return True
            elif str_input == "n":
                return False
            else:
                print(self.strfmr.get_colored_console_log("red",
                    "Error: Invalid input. Input 'Y'(=yes) or 'n'(=no)."))
    

class WinObserver(Observer):
    def __init__(self, uuid, is_alone):
        super().__init__(uuid, is_alone)
        from ganttlogger.modules.WinObservePackages import ActiveTabObserver
        self.ob_activetab = ActiveTabObserver(uuid, is_alone)


class MacObserver(Observer):
    def __init__(self, uuid, is_alone):
        super().__init__(uuid, is_alone)
        from ganttlogger.modules.MacObservePackages import ActiveTabObserver
        self.ob_activetab = ActiveTabObserver(uuid, is_alone)




import os
import time
from datetime import datetime
# import concurrent.futures as confu
import ganttlogger.modules.Global as global_v
from ganttlogger.modules.Public import StrFormatter

class Logger:
    uuid = ""
    strfmr = None
    dirname = ""
    def __init__(self, uuid):
        '''
        Get data from observer and log them.
        '''
        self.uuid = uuid
        self.strfmr = StrFormatter()
        # Create output dirs
        self.dirname = "ganttlogger_logs/{}".format(uuid)
        readme = "{}/README.txt".format(self.dirname)
        os.makedirs(os.path.dirname(readme), exist_ok=True)
        with open(readme, "w", encoding="utf-8") as f:
            startdate = datetime.now().strftime("%Y/%m/%d %H:%M:%S.%f")
            text = """\
UUID       : {uuid}
StartDate  : {startdate}
Supervisor : 
Target User: 
""".format(uuid=uuid, startdate=startdate)
            f.write(text)

    # def start(self):
    #     # Not implemented yet
    #     pass

    def run_logger(self):
        # Not implemented yet.
        # Run threads of receive_json() and output().
        print("Hello, Logger!")

    def receive_json(self):
        # Not implemented yet.
        # Recieve JSON and enqueue it.
        pass

    def output(self):
        with open("{}/active_tab.log".format(self.dirname), "a", encoding="utf-8") as ft,\
            open("{}/mouse.log".format(self.dirname), "a", encoding="utf-8") as fm,\
            open("{}/keyboard.log".format(self.dirname), "a", encoding="utf-8") as fk:
            # Write attributes
            ft.write("StartTime]:+:[ApplicationName]:+:[TabText\n")
            fm.write("Time]:+:[MoveDistance\n")
            fk.write("Time]:+:[PressCount\n")
            # Add logs
            is_all_empty = True
            while True:
                len_tab_queue = len(global_v.tab_queue)
                len_mouse_queue = len(global_v.mouse_queue)
                len_keyboard_queue = len(global_v.keyboard_queue)
                if len_tab_queue > 0:
                    is_all_empty = False
                    for _ in range(len_tab_queue):
                        data = global_v.tab_queue.popleft()
                        log = "{startTime}]:+:[{activeName}]:+:[{tabText}\n".format(
                            startTime=data["startTime"],
                            activeName=data["activeName"],
                            tabText=data["tabText"]
                        )
                        ft.write(log)
                if len_mouse_queue > 0:
                    is_all_empty = False
                    for _ in range(len_mouse_queue):
                        data = global_v.mouse_queue.popleft()
                        log = "{time}]:+:[{dis}\n".format(
                            time=data["time"],
                            dis=data["distance"]
                        )
                        fm.write(log)
                if len_keyboard_queue > 0:
                    is_all_empty = False
                    for _ in range(len_keyboard_queue):
                        data = global_v.keyboard_queue.popleft()
                        log = "{time}]:+:[{cnt}\n".format(
                            time=data["time"],
                            cnt=data["count"]
                        )
                        fk.write(log)
                if is_all_empty and global_v.cli_exit:
                    # This is a signal that all logging finished and CLI will exit
                    global_v.finish_logging = True
                    print(self.strfmr.get_colored_console_log("yellow",
                        "Logging all finished")) # maybe...
                    break
                if is_all_empty and (not global_v.is_sleeping):
                    # This is a signal that logging is finishing
                    print(self.strfmr.get_colored_console_log("yellow",
                        "All queue are empty"))
                is_all_empty = True
                time.sleep(2) # It's OK any seconds





from collections import deque

'''
Variables work for managing a flag whether observers's threads should exit.
'''
# Whether pausing children threads? (But actually, working in the backend)
is_sleeping = False
# Whether thread-loop exit because of error catch?
is_threadloop_error = False
# Whether threads 'Observer' and 'Logger' exited?
cli_exit = False
# Whether 'Logger' has finished all logging?
finish_logging = False


# Data structure is based on JSON.
'''
{
    "uuid": "",
    "type": "t",
    "id": -1,
    "activeName": "",
    "tabText": "",
    "startTime": ""
}
'''
tab_queue = deque([])
tab_id = -1

'''
{
    "uuid": "",
    "type": "m",
    "id": -1,
    "distance": "",
    "time": ""
}
'''
mouse_queue = deque([])
mouse_id = -1

'''
{
    "uuid": "",
    "type": "k",
    "id": -1,
    "count": "",
    "time": ""
}
'''
keyboard_queue = deque([])
keyboard_id = -1


import sys
import platform
import uuid
from argparse import ArgumentParser, RawTextHelpFormatter
from ganttlogger.modules.Public import StrFormatter

class InitProcess:
    argparser = None
    strfmr = None
    def __init__(self):
        '''
        Prepare in some initial processes.
        '''
        self.argparser = ArgsParser()
        self.strfmr = StrFormatter()

    def get_init_parameters(self):
        '''
        Get initial parameters.

        Args:
            None

        Returns:
            os (str): OS and its version
            mode (str): "Alone" or "AloneWithPlot" or "Observer" or "Logger" or "Plotter" or "Displayer"
            uuid ("" | str): UUID when mode is not "Plotter" and "Displayer"
        '''
        os = self.get_os()
        mode = self.argparser.identify_mode()
        if (mode == "Plotter") or (mode == "Displayer"):
            uuid = "None"
        elif (mode == "Alone") or (mode == "Observer") or (mode=="AloneWithPlot"):
            uuid = self.generate_uuid()
        elif mode == "Logger":
            uuid = self.argparser.uuid
        else:
            print(self.strfmr.get_colored_console_log("red",
                "Error: Invalid variable in mode of InitProcess.py"))
            sys.exit()

        print(self.strfmr.get_colored_console_log("yellow", """\
OS        : {os}
mode      : {mode}
Your ID is: {uuid}
""".format(os=os, mode=mode, uuid=uuid)))

        # Get simpler os parameter
        if "Windows" in os:
            os = "w"
        elif "Darwin" in os:
            os = "d"

        return os, mode, uuid
    
    def get_os(self):
        '''
        Get OS and its version of platform.

        Args:
            None

        Returns:
            os (str): OS and its version
        '''
        os = platform.platform(terse=True)

        # This CLI can work on Windows or Mac
        if ("Windows" in os) or ("Darwin" in os):
            return os
        else:
            print(self.strfmr.get_colored_console_log("red",
                "Error: This can work on 'Windows' or 'MacOS'"))
            sys.exit()

    def generate_uuid(self):
        '''
        Generate UUID.

        Args:
            None

        Returns:
            (str): UUID
        '''
        return str(uuid.uuid4())

class ArgsParser:
    parser = None
    args = None
    uuid = ""
    strfmr = None
    def __init__(self):
        '''
        Parse cli options.
        '''
        self.strfmr = StrFormatter()
        usage = "ganttlogger [--observer] [--logger] [--uuid <UUID>] [--help] [--plotter] [--withplot] [--displayer]"
        self.parser = ArgumentParser(
            prog="ganttlogger",
            description="""\
This CLI will do Observing active-tab, mouse, keyboard,
and Logging them,
and Plotting graphs (active-tab=ganttchart, mouse=line, keyboard=bar).
{}""".format(self.strfmr.get_colored_console_log("yellow",
"If you don't set any option, this work both of 'observer' and 'logger'.")),
            usage=usage,
            formatter_class=RawTextHelpFormatter
        )
        self.parser.add_argument(
            "-o", "--observer",
            action="store_true",
            help="The role of this PC is only observing action."
        )
        self.parser.add_argument(
            "-l", "--logger",
            action="store_true",
            help="The role of this PC is only logging and plotting. You must also set '--uuid'."
        )
        self.parser.add_argument(
            "-u", "--uuid",
            type=str,
            dest="uuid",
            help="When you set '--logger', you must also set this by being informed from 'observer' PC."
        )
        self.parser.add_argument(
            "-p", "--plotter",
            action="store_true",
            help="Use this option if you want other outputs by a log in the current directory after getting one and a graph."
        )
        self.parser.add_argument(
            "--withplot",
            action="store_true",
            help="Use this option when you want to get a graph after running 'Alone'."
        )
        self.parser.add_argument(
            "-d", "--displayer",
            action="store_true",
            help="Use this option when you want to look a graph from a '.pkl' file."
        )
        self.args = self.parser.parse_args()

    def identify_mode(self):
        '''
        Identify mode from args.

        Args:
            None

        Returns:
            mode (str): "Alone" or "Observer" or "Logger"
        '''
        mode = ""
        if self.args.observer:
            mode = "Observer"
        elif self.args.logger:
            if not self.args.uuid:
                print(self.strfmr.get_colored_console_log("red",
                    "Error: Logger missing an option '--uuid <UUID>'."))
                sys.exit()
            mode = "Logger"
            self.uuid = self.args.uuid
        elif self.args.plotter:
            mode = "Plotter"
        elif self.args.displayer:
            mode = "Displayer"
        else:
            if self.args.uuid:
                print(self.strfmr.get_colored_console_log("red",
                    "Error: You may need '--logger'."))
                sys.exit()
            if self.args.withplot:
                mode = "AloneWithPlot"
            else:
                mode = "Alone"
        
        return mode


from datetime import datetime
import time
import psutil
from AppKit import NSWorkspace as nsw
from Quartz import (
    CGWindowListCopyWindowInfo,
    kCGWindowListOptionOnScreenOnly,
    kCGNullWindowID
)
import ganttlogger.modules.Global as global_v

class ActiveTabObserver:
    uuid = ""
    data_process = None
    def __init__(self, uuid, is_alone):
        '''
        Monitor active-tab on Mac.
        '''
        self.uuid = uuid
        if is_alone:
            self.data_process = self.enqueue_data
        else:
            self.data_process = self.send_json
    
    def run(self):
        try:
            recent_active_tab_text = "START!"
            while not global_v.is_sleeping:
                try:
                    fw = nsw.sharedWorkspace().activeApplication()
                    # active_pid = fw["NSApplicationProcessIdentifier"]
                    active_name = fw["NSApplicationName"]
                    active_tab_text = ""
                    cg_windows = CGWindowListCopyWindowInfo(kCGWindowListOptionOnScreenOnly, kCGNullWindowID)
                    for cg_window in cg_windows:
                        if active_name == cg_window["kCGWindowOwnerName"] and cg_window["kCGWindowName"]:
                            active_tab_text = cg_window["kCGWindowName"]
                            break
                except (KeyError, ValueError, psutil.NoSuchProcess):
                    # If not in time to get pid
                    # print("Warning: Failed in getting process information")
                    continue
                if recent_active_tab_text != active_tab_text.upper():
                    switched_time = datetime.now().strftime("%Y/%m/%d %H:%M:%S.%f")
                    recent_active_tab_text = active_tab_text.upper()
                    self.data_process(switched_time, active_name, active_tab_text)
                    # print("ActiveTab[{time}]: {pid}: {active_name}({tab_text})".format(
                    #     time=switched_time,
                    #     pid=active_pid,
                    #     active_name=active_name,
                    #     tab_text=active_tab_text))
                time.sleep(0.001)
            # Output the last log
            switched_time = datetime.now().strftime("%Y/%m/%d %H:%M:%S.%f")
            self.data_process(switched_time, active_name, active_tab_text)
        except:
            # If this thread stopped by rebooting from sleep, maybe...
            import traceback
            print("Thread loop exited by any problem!!!!")
            global_v.is_threadloop_error = True
            global_v.is_sleeping = True
            traceback.print_exc()
        # except KeyboardInterrupt:
        #     print("ActiveTabObserver.py: KeyboardInterrupt")

    def send_json(self, t, active_name, tab_text):
        pass

    def enqueue_data(self, t, active_name, tab_text):
        global_v.tab_id += 1
        global_v.tab_queue.append({
            "uuid": self.uuid,
            "type": "t",
            "id": global_v.tab_id,
            "activeName": active_name,
            "tabText": tab_text,
            "startTime": t
        })



'''
# If neccesary to implement observer for each OS, implement below.
class MouseObserver:
    pass
class KeyboardObserver:
    pass
'''